--[[Author:		DieenDieen 
-- global lookup
]]

local folder, core = ...
TaR = core --global table

-- local
title		= "TaR"
version		= GetAddOnMetadata(folder, "X-Curse-Packaged-Version") or ""
titleFull	= title.." "..version
packedtitle = "TaR";


TaRHistory={};
TaRRun=0;

local TaRFrame = false;
local TrackingEnabled = false;
-- Isolate the environment

core = LibStub("AceAddon-3.0"):NewAddon(packedtitle, "AceConsole-3.0",  "AceHook-3.0", "AceEvent-3.0","AceSerializer-3.0","AceComm-3.0") 

local CommPrefix="TaRmsg";
local librc = LibStub("LibRangeCheck-2.0");
local libc = LibStub:GetLibrary("LibCompress");

local InteractDistances={28;11.11;9.9;28};
if librc==nil then
	librc={};
   function librc:GetRange(unit)
   -- simplified replacement if LibRangeCheck is not installed
   	local MinRange,MaxRange;
		for i=1,4 do
				if CheckInteractDistance(unit,i) then 
					MaxRange=math.min(MaxRange or 999999,InteractDistances[i]);
					MinRange=math.max(MinRange or 0,0);
					else
					MinRange=math.max(MinRange or 0,InteractDistances[i]);
					end;
			end;
		if UnitInRange(unit) then
			MaxRange=math.min(MaxRange or 999999,40);
			MinRange=math.max(MinRange or 0,0);
			else
			MinRange=math.max(MinRange or 0,40);
			end;
		if	UnitIsVisible(unit) then
			MaxRange=math.min(MaxRange or 999999,100);
			MinRange=math.max(MinRange or 0,0);
			else
			MinRange=math.max(MinRange or 0,100);
			end;
		return MinRange,MaxRange;
   	end;
   end;

local regEvents = {
	"ADDON_LOADED",
	"UI_INFO_MESSAGE",
	"ZONE_CHANGED",
	"ZONE_CHANGED_INDOORS",
	"ZONE_CHANGED_NEW_AREA",
	"ACTIVE_TALENT_GROUP_CHANGED",
	"CHARACTER_POINTS_CHANGED",
	"PLAYER_TALENT_UPDATE",
}

 
function CompareTablesByValue (first,second)
if type(first) == "table" and type(second) == "table" then
		for key, value in pairs (first) do
			if type(value) == "table" and not CompareTablesByValue (value,second[key])  then return false; end;
         if type(value) ~= "table" and second[key]~=value then return false; end;            			
		end;
		for key, value in pairs (second) do
			if type(value) == "table" and not CompareTablesByValue (value,first[key])  then return false; end;
         if type(value) ~= "table" and first[key]~=value then return false; end;            			
		end;
		return true;
	else
	return false;
	end;
end;

local function GetMinDurability()
	MinDurability=1;
	for i=1,18 do
		local current, maximum = GetInventoryItemDurability(i);		 
		if current~=nil and  maximum~=nil and current<maximum then MinDurability=math.min (MinDurability,current/maximum);end;
		end;
	return MinDurability;
end;


local function PackPart (Actor,Guid,Part,Events,LastEvent)
	local res=Actor[Part];
	if type(res) == "table" then
		if next(res) then
			if LastEvent>0 and Events[LastEvent].Actors[Guid] then 
				local LEP=Events[LastEvent].Actors[Guid][Part];
				while type(LEP) == "table" and type(LEP.CopyOf)=="number" do LastEvent=LEP.CopyOf; LEP=Events[LastEvent].Actors[Guid][Part]; end;
				if type(LEP) == "table" then
					if CompareTablesByValue (res,LEP) then res={CopyOf=LastEvent}; end;			
				end;			
			end;
		else
			res = nil;
		end;
	end;
	return res;
end;

local TalentSpecid, TalentSpecname, TalentSpecpointsSpent;

local function FindActiveTalentSpec ()
   TalentSpecpointsSpent=0;
   for i = 1, GetNumTalentTabs() do
      local lTalentSpecid, lTalentSpecname, _, _, lTalentSpecpointsSpent= GetTalentTabInfo( i, false, false, nil );
      if  type(lTalentSpecpointsSpent)=="number" and lTalentSpecpointsSpent>TalentSpecpointsSpent then  
         TalentSpecid, TalentSpecname, TalentSpecpointsSpent = lTalentSpecid, lTalentSpecname, lTalentSpecpointsSpent;
      end;
   end;
end;

local function GetActiveTalentSpec ()
   if not TalentSpecid then FindActiveTalentSpec (); end;
   return TalentSpecid, TalentSpecname, TalentSpecpointsSpent;
end

function StoreTrackedUnit (Actors,Guids,Target,TargetType)
	if UnitExists(Target) and UnitIsUnit(Target,Target) then
		local guid = UnitGUID(Target);
		local unitX, unitY = GetPlayerMapPosition(Target);
		local unitclass, classFilename, unitrace, raceFilename, unitsex, unitname, unitrealm = GetPlayerInfoByGUID(guid);
		local unitspeed, groundSpeed, flightSpeed, swimSpeed = GetUnitSpeed(Target);
		if not unitrealm or unitrealm=="" then unitrealm = GetRealmName();end;
		
		
		if type(Guids[guid]) ~= "table" then
		  -- new guid encountered
			Guids[guid] = {};
			local G= Guids[guid];
			G.Guid = guid;			
			G.UnitType = tonumber(guid:sub(5,5), 16) % 8;
			if G.UnitType==3 then -- NPC			   
			   G.NPCID = tonumber(strsub(guid,7,10),16); 
			   end;
			G.Faction = UnitFactionGroup(Target);
			G.Class = unitclass;
			local unitclass, unitclassFileName = UnitClass(Target);
			G.Class = unitclass;G.DefaultColor=RAID_CLASS_COLORS[unitclassFileName];
			G.Realm = unitrealm; 
			G.Sex = unitsex;
			if unitname and unitname~="" then G.Name = unitname;else G.Name = UnitName(Target);end;
			G.Race = unitrace;			
			G.Classification,G.CreatureType,G.CreatureFamily = UnitClassification(Target),UnitCreatureType(Target),UnitCreatureFamily(Target);
			G.Player = UnitPlayerControlled(Target);
			
			local unitcanBeTank, unitcanBeHealer, unitcanBeDPS = UnitGetAvailableRoles(Target);
			G.canBeTank,G.canBeHealer,G.canBeDPS = unitcanBeTank or nil,unitcanBeHealer or nil,unitcanBeDPS or nil; 			
			
			local unitlevel = UnitLevel(Target);
			if unitlevel>0 then G.Level = unitlevel end;
			local UnitGuild, UnitGuildRankName, UnitGuildRankIndex = GetGuildInfo(Target);
			if UnitGuild then G.Guild, G.GuildRankName, G.GuildRankIndex = UnitGuild, UnitGuildRankName, UnitGuildRankIndex;end;
			end;
		
		local G= Guids[guid];
	
		if type(Actors[guid]) ~= "table" then Actors[guid] = {}; end;	local CE = Actors[guid];
		
		if type(CE.Zone) ~= "table" then CE.Zone = {}; end; local CEZ=CE.Zone;
		if type(CE.Coords) ~= "table" then CE.Coords = {}; end;local CEC=CE.Coords;
		if type(CE.Target) ~= "table" then CE.Target = {}; end;local CET=CE.Target;
		if type(CE.Status) ~= "table" then CE.Status = {}; end;local CES=CE.Status;
		if type(CE.Energies) ~= "table" then CE.Energies = {}; end;	local CEE = CE.Energies;
		if type(CE.Status) ~= "table" then CE.Status = {}; end;local CES=CE.Status;
		if type(CE.Pet) ~= "table" then CE.Pet = {}; end;local CEP=CE.Pet;		
		
		if  type(CE.PovLevel) ~= "number" then CE.PovLevel = 0; end; 
		if CE.PovLevel<1 then
			CE.PovLevel = 1;
			CE.Guid = guid;
			
			
			if (unitX or 0)>0 or (unitY or 0)>0 then CEC.X,CEC.Y = unitX, unitY;end;
			CEC.Speed = unitspeed;
			
						 
			local unitTarget=UnitGUID(Target.."target");
			if unitTarget then      
				CET.Guid = unitTarget;
				CET.Friendly = UnitIsFriend(Target, Target.."target");
				local isTanking, threatstatus, threatpct, rawthreatpct, threatvalue = UnitDetailedThreatSituation(Target, Target.."target");
				CET.Threat=threatstatus;				
				end;
				
			local unitPet=UnitGUID(Target.."pet");
			if unitPet then      
				CEP.Guid = unitPet;
				CEP.Friendly = UnitIsFriend(Target, Target.."pet");								
				end;				
						  			
			if not UnitIsConnected(Target) then CES.NotConnected = true;end;
			CES.Corpse = UnitIsCorpse(Target);
			CES.Dead,CES.Ghost = UnitIsDead(Target),UnitIsGhost(Target);
			CES.FeignDeath = UnitIsFeignDeath(Target);
			CES.UsingVehicle = UnitUsingVehicle(Target);
			CES.Charmed = UnitIsCharmed(Target);
			CES.Possessed = UnitIsPossessed(Target);
			CES.AFK = UnitIsAFK(Target);
			CES.Marker = GetRaidTargetIndex(Target);
			local unitrole=UnitGroupRolesAssigned(Target);
			if unitrole and unitrole~="NONE" then CES.Role = unitrole;end;
			CES.InCombat = UnitAffectingCombat(Target); 
			
									
			CEE.Type = UnitPowerType(Target);
			CEE.Power,CEE.PowerMax = UnitPower(Target,CEE.Type),UnitPowerMax(Target,CEE.Type);
			CEE.Health,CEE.HealthMax = UnitHealth(Target),UnitHealthMax(Target);
			CEE.SpellHaste = UnitSpellHaste(Target);			
			 
			local CEA = {};
			for i=1,32 do 
				local  auraname, aurarank, auraicon, auracount, auradispelType, auraduration, auraexpires, auracaster, auraisStealable, aurashouldConsolidate, auraspellID, auracanApplyAura, auraisBossDebuff, auravalue1, auravalue2, auravalue3 = UnitAura(Target, i);
				if auraname and auraname ~= "" then 
					local aura = {};
					if aurarank~= "" then aura.Rank = aurarank;end;
					if auracount>0 then aura.Count = auracount;end;
					if auraexpires>0 then aura.Expires = auraexpires;end;
					aura.Caster = UnitGUID(auracaster or "") or auracaster;
					aura.SpellID = auraspellID;
					aura.BossDebuff = auraisBossDebuff;
					CEA[#CEA+1]=aura;
					end; 
				end; 
	 
				   
			CE.Auras = CEA;
			end;			
		  
		if TargetType=="player" then
			CE.PovLevel = 10;
					
			CEC.Facing = GetPlayerFacing();									
			CES.HasFullControl = HasFullControl();
			CES.Falling = IsFalling();
			CES.Mounted = IsMounted();
			CES.Stealthed = IsStealthed();
			CES.Swimming = IsSwimming();
			CES.MinDurability=GetMinDurability();
			CES.Language= GetDefaultLanguage();			

			CES.TalentSpecid,CES.TalentSpecName = GetActiveTalentSpec ();
			
			
			CEZ.MapAreaID = GetCurrentMapAreaID();
			CEZ.MapLevel = GetCurrentMapDungeonLevel();
			CEZ.ZoneName = GetRealZoneText();
			CEZ.SubZoneName = GetSubZoneText();
			
			local iname, itype, idifficultyIndex, idifficultyName, imaxPlayers, idynamicDifficulty, isDynamic = GetInstanceInfo();
			if itype and itype ~= "none" then
				CEZ.Type = itype;
				CEZ.Difficulty = idifficultyName;
				end;

			if CET.Guid then
			   CET.MinDistance,CET.MaxDistance=librc:GetRange(Target.."target");				
				end;
			if CEP.Guid then
			   CEP.MinDistance,CEP.MaxDistance=librc:GetRange(Target.."pet");				
				end;
						
		end;
		end;
end;


function StoreTrackedUnitPT (Actors,Guids,Target,TargetType)
   StoreTrackedUnit (Actors,Guids,Target,TargetType);
   StoreTrackedUnit (Actors,Guids,Target .."target","");
   StoreTrackedUnit (Actors,Guids,Target .."targettarget","");   
   StoreTrackedUnit (Actors,Guids,Target .."pet","pet");
   StoreTrackedUnit (Actors,Guids,Target .."pettarget","");
   StoreTrackedUnit (Actors,Guids,Target .."pettargettarget","");
end;

local LastEvent=0;
function StoreTrackedEvent (TaRHistory)
	local CurrentEvent={Infos={},Actors={}};
	local Actors=CurrentEvent.Actors;
	local Infos=CurrentEvent.Infos;
	local Events=TaRHistory.Events;
	local Guids=TaRHistory.Guids;	
	
	Infos.Time = GetTime();
	
	StoreTrackedUnitPT (Actors,Guids,"player","player");
	StoreTrackedUnitPT (Actors,Guids,"pet","pet");
	StoreTrackedUnitPT (Actors,Guids,"focus","");
	StoreTrackedUnitPT (Actors,Guids,"npc","");
	
	for i=1,GetNumRaidMembers() do
		StoreTrackedUnitPT (Actors,Guids,"raid" .. i,"friendly");   
		end
	
	for i=1,GetNumPartyMembers() do
		StoreTrackedUnitPT (Actors,Guids,"party" .. i,"friendly");
		end
	
	for i=1,4 do
		StoreTrackedUnitPT (Actors,Guids,"boss" .. i,"boss");	
		end;		
	if true then --simple packing
	   for Guid, Guidinfo in pairs (Actors) do
	      if type(Guidinfo) == "table" then
				for Part, value in pairs (Guidinfo) do
					if type(value) == "table" then Actors[Guid][Part] = PackPart (Actors[Guid],Guid,Part,Events,LastEvent);end;
				end;
			end;
		end;
	end;
		 tinsert(Events,CurrentEvent);
		 LastEvent = #Events;
end

local UpdateTimerTotal = 0;
local UpdateRunning=false;
local PlannedRunTime=0;

local ScanInterval=0.1;


function TaRonUpdate(self,elapsed)
   if TrackingEnabled then
       UpdateTimerTotal = UpdateTimerTotal + elapsed;
       --if not UpdateRunning and UpdateTimerTotal >= ScanInterval then
       	if not UpdateRunning and GetTime() >= PlannedRunTime then
           UpdateRunning = true;
           UpdateTimerTotal = 0;           
           local WorkStartTime = GetTime();
           PlannedRunTime=math.ceil((WorkStartTime+ScanInterval)*100)/100;
           if type(TaRHistory[TaRRun]) ~= "table" then TaRHistory[TaRRun] = {TotalTime=0;TotalCount=0;TotalWorkTime=0;Events={};Guids={}}; end;
           local TaRH = TaRHistory[TaRRun]; 		    
		   TaRH.TotalTime=TaRH.TotalTime+UpdateTimerTotal;
		   TaRH.TotalCount=TaRH.TotalCount+1;
		   		   
		   StoreTrackedEvent (TaRH);
		   TaRH.TotalWorkTime=GetTime()-WorkStartTime +TaRH.TotalWorkTime;		   
		   UpdateRunning = false;
		end;
     end;
end

function core:OnInitialize()
    self:RegisterChatCommand("tarcmd", "MySlashProcessorFunc");
    TaRRun = TaRRun + 1;
    print ("tar started");	
end

function core:OnEnable()
   for i, event in pairs (regEvents) do 
		self:RegisterEvent(event)
	end	
end

function core:ADDON_LOADED(event, ...)
	core:RegisterComm(CommPrefix);
    if not TaRFrame then
       TaRFrame = CreateFrame("frame");       
       end    
end

function core:OnCommReceived(prefix, message, distribution, sender)
    print (prefix, message, distribution, sender);
    local remotetime=tonumber(message);
    print (GetTime(),remotetime);
end;

function core:ZONE_CHANGED(event, ...)
   SetMapToCurrentZone();
end

function core:ZONE_CHANGED_INDOORS(event, ...)
   SetMapToCurrentZone();
end

function core:ZONE_CHANGED_NEW_AREA(event, ...)
   SetMapToCurrentZone();
end

function core:ACTIVE_TALENT_GROUP_CHANGED(event, ...)
	TalentSpecid=nil;
end;

function core:CHARACTER_POINTS_CHANGED(event, ...)
	TalentSpecid=nil;
end;

function core:PLAYER_TALENT_UPDATE(event, ...)
	TalentSpecid=nil;
end;

function core:UI_INFO_MESSAGE(event, ...)
local arg1 = ...;
end;


function core:OnDisable()
	
end


----------------------------------------------
function core:MySlashProcessorFunc(input)	--
----------------------------------------------
    if input=="" then
      print ("--- TaR "..version.." is running");
      print ("use start or stop to control tracking state");
      UpdateAddOnMemoryUsage();
      print ("atm uses "..GetAddOnMemoryUsage("TaR").." kb of memory");
      elseif strlower(input)=="continue" then TrackingEnabled=true;TaRFrame:SetScript("OnUpdate", TaRonUpdate);
      elseif strlower(input)=="start" then TrackingEnabled=true;TaRRun=TaRRun+1; TaRFrame:SetScript("OnUpdate", TaRonUpdate);
      elseif strlower(input)=="stop" then TrackingEnabled=false;TaRFrame:SetScript("OnUpdate", nil);
      elseif strlower(input)=="beep" then core:SendCommMessage(CommPrefix, tostring(GetTime()), "GUILD");
      elseif strlower(input)=="compress" then 
         TrackingEnabled=false;TaRFrame:SetScript("OnUpdate", nil);
         TaRHistory[TaRRun] = {serialized=core:Serialize(TaRHistory[TaRRun])};
         if libc then TaRHistory[TaRRun] = {compressed=libc:CompressHuffman(TaRHistory[TaRRun].serialized)};end;
    end;
   if TrackingEnabled then
     print ("--- Tracking running");
     else
     print ("--- Tracking stopped");
     end;
end



